;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - preview.ss                                                     #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; This file contains general functions used in previewing.

;; Deletes an element in the preview
(define (preview-delete-element element next-data)
  ;; Is the element a frame?
  (if (eq? (car element) 'frame)
      ;; Then we have no choice, then - just hide it.
      (send (get-property-value element 'pointer) show #f)
      ;; Otherwise, we need to check whether we can update the preview...
      (if (member (car element) '(pane horizontal-pane vertical-pane))
          ;; We can't without reloading the whole project!
          (let ((reload-result (message-box/custom "Warning"
                                            (string-append "You are about to delete a non-subwindow object. The preview will not reflect this change"
                                                           " until you reload the project (i.e. choose 'Reload the project', or close and reopen the project). "
                                                           "What do you want to do?")
                                            "Reload the project" "Don't update the preview" #f hierarchy-frame '(default=1 number-order caution) 2)))
            
            ;; Ok, the user wants to reload the project...
            (if (= 1 reload-result)
                ;; ...then obey him ;)
                (begin (set! project-data-global next-data)
                       (preview-update-preview next-data next-data))))

          ;; Otherwise, we get a pointer to the parent...
          (send (get-property-value (project-data-get-parent element) 'pointer)
                ;; To which we ask to delete a child...
                delete-child
                ;; Whose pointer is obtained with the element line...
                (get-property-value element 'pointer)))))

;; This function is called for example when the user opens a new file : it
;; then removes the preview from the previous file and creates the preview
;; from the new file...
;; It is also called when the user updates the preview "on the fly" (when
;; the user changes the style property of a container).
(define (preview-update-preview old-data new-data)
  ;; First we develop a little function which will allow us to simply
  ;; hide (because there is no way to delete them) frames.
  (define (hide-if-frame line)
    (if (eq? (car line) 'frame)
        (send (get-property-value line 'pointer) show #f)))
  ;; Then we apply the function on each element of the list.
  (apply-procedure-on-data-list old-data hide-if-frame)
  
  ;; Ok, we have hidden all the widgets from the old-data. Now we are going
  ;; to parse the new list, and for each line, we will first create the associated
  ;; widget, and then update the line so that the pointer to the widget is inserted
  ;; into it. This might be the most difficult part of this project - but I will try
  ;; to comment it as much as possible.
  
  ;; First we create a parser that will take a line from the new data, and computes it
  ;; so that it can be passed to new, in order to create the widget
  (letrec ((parse-list (lambda (line parent)
                         (cond ((null? line) line)
                               ;; We don't want (necessary) nor (optional)
                               ((symbol? (car line)) (parse-list (cdr line) parent))
                               ;; All these tags are not properties
                               ((eq? (caar line) 'children) (parse-list (cdr line) parent))
                               ((eq? (caar line) 'necessary) (parse-list (cdr line) parent))
                               ((eq? (caar line) 'optional) (parse-list (cdr line) parent))
                               ((eq? (caar line) 'selected) (parse-list (cdr line) parent))
                               ((eq? (caar line) 'collapsed) (parse-list (cdr line) parent))
                               ((eq? (caar line) 'id) (parse-list (cdr line) parent))
                               ;; This one was added to be able to reload the project "on the fly".
                               ((eq? (caar line) 'pointer) (parse-list (cdr line) parent))
                               ;; We need to update the "parent" property
                               ((eq? (caar line) 'parent) (cons (list 'parent parent) (parse-list (cdr line) parent)))
                               ;; If this is a valid property, continue with the remaining list...
                               (else (cons (car line) (parse-list (cdr line) parent))))))
           
           ;; Then we develop a function that will take a line from the new data, call the necessary functions to preview it,
           ;; then return two results : the computed line and the pointer to the newly created widget.
           (line-helper (lambda (line parent)
                          ;; We ignore the "Project" line...
                          (cond ((eq? (car line) 'project) (cons line #f))
                                ;; For each line, we call 'new' with the right arguments
                                (else (let ((pointer (eval (append (list 'new) 
                                                                   (list (string->symbol (string-append (symbol->string (car line)) "%")))
                                                                   (parse-list line parent)))))
                                        ;; If the line is a window, we need to show it!
                                        (if (eq? (car line) 'frame)
                                            (send pointer show #t))
                                        ;; And then, we send the two needed results.
                                        (cons (append (remove-pointer-if-any line) (list (list 'pointer pointer))) pointer))))))
           
           ;; This function removes the pointer property for fields created on the fly so that it can be replaced by the new pointer value.
           (remove-pointer-if-any (lambda (line)
                                    ;; First we get the pointer list...
                                    (let ((pointer-list (assq 'pointer (cdr line))))
                                      ;; If there is one...
                                      (if pointer-list
                                          ;; ...we just remove it...
                                          (cons (car line) (remq pointer-list (cdr line)))
                                          ;; ...otherwise, we have nothing to do.
                                          line))))
           
           ;; Then, the main loop, which computes the whole data list - mainly inspired from the apply-function-on-data-list function,
           ;; but slightly more complex, because we have to deal with the widgets' parents.
           (list-helper (lambda (data-list parent)
                          ;; If the data-list is empty, the job is done!
                          (if (null? data-list)
                              data-list
                              (begin
                                ;; Otherwise, we apply line-helper on the line...
                                (let* ((results (line-helper (car data-list) parent))
                                       ;; And dispatch the results...
                                       (mod-current-ptr (cdr results))
                                       (mod-current-line (car results))
                                       
                                       ;; First recursive call (for children)... To which we pass mod-current-ptr because
                                       ;; the pointer to their parent is the pointer to the current element...
                                       (mod-current-children (if (assq 'children (cdar data-list)) 
                                                                 (list-helper (cdr (assq 'children (cdr mod-current-line))) 
                                                                              mod-current-ptr) #f)))
                                  
                                  ;; We build the result and call the second recursive call (for brothers)... To which we
                                  ;; pass the current parent because our brother and us have the same parents!
                                  (cons (if mod-current-children
                                            (set-children mod-current-line mod-current-children)
                                            mod-current-line)
                                        (list-helper (cdr data-list) parent))))))))
    
    ;; And then we call list-helper, and update project-data-global with the result...
    (set! project-data-global (list-helper new-data #f))))

;; This function updates, in the preview, the order of the element's children
(define (preview-update-children-order element)
  ;; This filter returns the updated children list...
  (define (widget-filter widget-list)
    ;; Just get the children list and compute it into a list of pointers...
    (map (lambda (elt) (get-property-value elt 'pointer)) (filter (lambda (elt) (not (eq? (car elt) 'frame))) (get-children element))))
      
  ;; We ask the widget to update its children using the filter we just wrote...
  (send (get-property-value element 'pointer) change-children widget-filter))

;; This function is called in the properties-widget on update-out.
(define (preview-update-widget widget-line convert-elt last-value-changed) 
  ;; If the widget is a container and the last value changed is a property that we
  ;; cannot change "on the fly", then, we build the whole widget again!
  (if (and (not (container? widget-line))
           (or (eq? 'choices last-value-changed)
               (eq? 'label last-value-changed)
               (eq? 'callback last-value-changed)
               (eq? 'style last-value-changed)
               (eq? 'min-value last-value-changed)
               (eq? 'max-value last-value-changed)))
      ;; We will have to save a pointer to the newly created element...
      (let ((new-pointer '?))
        ;; This parser allows us to compute the line in a call to "new"...
        (define (parse-list L)
          (cond ((null? L) '())
                ((eq? (caar L) 'parent)(cons (list 'parent (get-property-value (project-data-get-parent widget-line) 'pointer));;!!
                                             (parse-list (cdr L))))
                ((eq? (caar L) 'necessary)(parse-list (cdr L)))
                ((eq? (caar L) 'optional)(parse-list (cdr L)))
                ((eq? (caar L) 'selected)(parse-list (cdr L)))
                ((eq? (caar L) 'collapsed)(parse-list (cdr L)))
                ((eq? (caar L) 'pointer)(parse-list (cdr L)))
                ((eq? (caar L) 'children) (parse-list (cdr L)))
                ((eq? (caar L) 'id)(parse-list (cdr L)))
                ((eq? (cadar L) 'na) (parse-list (cdr L)))
                (else (cons (car L) (parse-list (cdr L))))))
        ;; This function parses the parent's children list... Looking for the current widget.
        (define (filter widget-list)
          ;; If the list is empty, then we have nothing to do...
          (cond ((null? widget-list) widget-list)
                ;; Did we find the current element?
                ((eq? (car widget-list) (get-property-value widget-line 'pointer))
                 ;; Yes we did - therefore we instantiate it again...
                 (let ((new-element (eval `(new ,(symbol-append (car widget-line) '%)
                                                ,@(parse-list (cdr widget-line))))))
                   ;; ...save the pointer...
                   (set! new-pointer new-element)
                   ;; ...and return the updated children list.
                   (cons new-element (cdr widget-list))))
                ;; Otherwise, we do not change the currently parsed element.
                (else (cons (car widget-list) (filter (cdr widget-list))))))
        
        ;; This function updates the current widget line, setting the right pointer for the pointer property.
        (define (update-line line)
          ;; If the line is empty, do nothing.
          (cond ((null? line) line)
                ;; If the first element is a symbol, just keep it.
                ((symbol? (car line))(cons (car line) (update-line (cdr line))))
                ;; If the first element is the pointer, update it.
                ((eq? (caar line) 'pointer) (cons (list 'pointer new-pointer) (update-line (cdr line))))
                ;; Otherwise, just keep the first element.
                (else (cons (car line) (update-line (cdr line))))))
        
        ;; We ask the widget's parent...
        (send (get-property-value (project-data-get-parent widget-line) 'pointer)
              ;; ...to update its children.
              change-children filter)

        ;; And then return the updated line.
        (update-line widget-line))
      
      ;; Otherwise, we will just change each property separately.
      (begin
        ;; Returns the field list, given the field name
        (letrec ((get-field (lambda (name)
                              (assq name (cdr widget-line)))))
          ;; Assign the widget pointer to a variable
          (let ((widget-pointer (cadr (get-field 'pointer))))
            ;;
            ;; And then we have (almost) no choice, because MrEd's function names differ,
            ;; and there is no real rule, therefore, we have to enumerate each case!
            ;; (We could, of course, make the code a bit shorter but it wouldn't be
            ;; worth doing it).
            ;;
            (case last-value-changed
              ;; These ones begin with "set"...
              ((label) (send widget-pointer set-label (cadr (convert-elt (get-field 'label)))))
              ((selection) (send widget-pointer set-selection (cadr (convert-elt (get-field 'selection)))))
              ((value) (send widget-pointer set-value (cadr (convert-elt (get-field 'value)))))
              ((init-value) (send widget-pointer set-value (cadr (convert-elt (get-field 'init-value)))))
              ((range) (send widget-pointer set-range (cadr (convert-elt (get-field 'range)))))
              
              ;; These ones have no prefix...
              ((horiz-margin) (send widget-pointer horiz-margin (cadr (convert-elt (get-field 'horiz-margin)))))
              ((min-width) (send widget-pointer min-width (cadr (convert-elt (get-field 'min-width)))))
              ((min-height) (send widget-pointer min-height (cadr (convert-elt (get-field 'min-height)))))
              ((vert-margin) (send widget-pointer vert-margin (cadr (convert-elt (get-field 'vert-margin)))))
              ((stretchable-width) (send widget-pointer stretchable-width (cadr (convert-elt (get-field 'stretchable-width)))))
              ((stretchable-height) (send widget-pointer stretchable-height (cadr (convert-elt (get-field 'stretchable-height)))))
              ((vert-margin) (send widget-pointer vert-margin (cadr (convert-elt (get-field 'vert-margin)))))
              ((horiz-margin) (send widget-pointer horiz-margin (cadr (convert-elt (get-field 'horiz-margin)))))
              ((border) (send widget-pointer border (cadr (convert-elt (get-field 'border)))))
              ((spacing) (send widget-pointer spacing (cadr (convert-elt (get-field 'spacing)))))
              
              ;; These ones are really weird !
              ((width) (send widget-pointer resize (cadr (convert-elt (get-field 'width))) (send widget-pointer get-height)))
              ((height) (send widget-pointer resize (send widget-pointer get-width) (cadr (convert-elt (get-field 'height)))))
              ((x) (send widget-pointer move (cadr (convert-elt (get-field 'x))) (send widget-pointer get-y)))
              ((y) (send widget-pointer move (send widget-pointer get-x) (cadr (convert-elt (get-field 'y)))))
              ((enabled) (send widget-pointer enable (cadr (convert-elt (get-field 'enabled)))))
              ((alignment) (send widget-pointer set-alignment (car (eval (cadr (convert-elt (get-field 'alignment)))))
                                 (cadr (eval (cadr (convert-elt (get-field 'alignment))))))))
           
            ;; We warn the parent that the flow has been modified
            ;(let ((parent-pointer (project-data-get-parent widget-line)))
            ;  (if parent-pointer
            ;      (send (get-property-value parent-pointer 'pointer) container-flow-modified)))
            
            ;; Then we return the unchanged line...
            widget-line)))))